package ij.measure;
import ij.IJ;

/** This is a table for storing measurement results as columns of real numbers. */
public class ResultsTable {

	public static final int MAX_COLUMNS = 50;
	
	public static final int AREA=0, MEAN=1, STD_DEV=2, MODE=3, MIN=4, MAX=5,
		X_CENTROID=6, Y_CENTROID=7, X_CENTER_OF_MASS=8, Y_CENTER_OF_MASS=9,
		PERIMETER=10, ROI_X=11, ROI_Y=12, ROI_WIDTH=13, ROI_HEIGHT=14,
		USER1=15, USER2=16, USER3=17, USER4=18;

	private String[] headings = {"Area","Mean","StdDev","Mode","Min","Max",
		"X","Y","XM","YM","Perim.","BX","BY","Width","Height",
		"User1", "User2", "User3", "User4"};
	private int counter;
	private float[][] columns = new float[MAX_COLUMNS][];
	private int maxRows = 100; // will be increased as needed
	private int lastColumn = -1;
	private	StringBuffer sb;
	private int precision = 3;

	/** Constructs an empty ResultsTable with the counter=0 and no columns. */
	public ResultsTable() {
	}
	
	/** Increments the measurement counter by one. */
	public synchronized void incrementCounter() {
		counter++;
		if (counter==maxRows) {
			for (int i=0; i<MAX_COLUMNS; i++) {
				if (columns[i]!=null) {
					float[] tmp = new float[maxRows*2];
					System.arraycopy(columns[i], 0, tmp, 0, maxRows);
					columns[i] = tmp;
				}
			}
			maxRows *= 2;
		}
	}
	
	/** Returns the current value of the measurement counter. */
	public int getCounter() {
		return counter;
	}
	
	/** Sets the last value in the specified column. Counter must be >0.*/
	public void addValue(int column, double value) {
		if ((column<0) || (column>=MAX_COLUMNS))
			throw new IllegalArgumentException("Index out of range: "+column);
		if (counter==0)
			throw new IllegalArgumentException("Counter==0");
		if (columns[column]==null) {
			columns[column] = new float[maxRows];
			if (column>lastColumn) lastColumn = column;
		}
		columns[column][counter-1] = (float)value;
	}
	
	/** Returns a copy of the specified column as a float array.
		Returns null if the column is empty. */
	public float[] getColumn(int column) {
		if ((column<0) || (column>=MAX_COLUMNS))
			throw new IllegalArgumentException("Index out of range: "+column);
		if (columns[column]==null)
			return null;
		else {
			float[] data = new float[counter];
			for (int i=0; i<counter; i++)
				data[i] = columns[column][i];
			return data;
		}
	}
	
	/** Returns a tab-delimited string containing the column headings. */
	public String getColumnHeadings() {
		StringBuffer sb = new StringBuffer(200);
		sb.append(" \t");
		for (int i=0; i<=lastColumn; i++) {
			if (columns[i]!=null)
				sb.append(headings[i] + "\t");
		}
		return new String(sb);
	}

	/** Returns a tab-delimited string representing the
		specified row, where 1<=row<=counter. */
	public String getRowAsString(int row) {
		if ((row<1) || (row>counter))
			throw new IllegalArgumentException("Row out of range: "+row);
		if (sb==null)
			sb = new StringBuffer(200);
		else
			sb.setLength(0);
		sb.append(Integer.toString(row));
		sb.append("\t");
		for (int i=0; i<=lastColumn; i++) {
			if (columns[i]!=null)
				sb.append(n(columns[i][row-1]));
		}
		return new String(sb);
	}
	
	/** Changes the heading of the specified column. */
	public void setHeading(int column, String heading) {
		if ((column<0) || (column>=headings.length))
			throw new IllegalArgumentException("Column out of range: "+column);
		headings[column] = heading;
	}
	
	/** Sets the number of digits to the right of decimal point. */
	public void setPrecision(int precision) {
		this.precision = precision;
	}
	
	String n(double n) {
		String s;
		if (Math.round(n)==n)
			s = IJ.d2s(n,0);
		else
			s = IJ.d2s(n,precision);
		return s+"\t";
	}
		
	/** Clears all the columns and sets the counter to zero. */
	public synchronized void reset() {
		counter = 0;
		maxRows = 100;
		lastColumn = -1;
		for (int i=0; i<MAX_COLUMNS; i++)
			columns[i] = null;
	}
	
	public String toString() {
		return ("ctr="+counter);
	}
	
}
