package ij.process;

import java.util.*;
import java.awt.*;
import java.awt.image.*; 
import ij.gui.*;

/** This abstract class is the superclass for classes that process
	the four data types (byte, short, float and RGB) supported by ImageJ.
*/

public abstract class ImageProcessor extends Object {

	/** Value of pixels included in masks. */
	public static final int BLACK = 0xFF000000;
	
	/** Value returned by getMinThreshold() when thresholding is not enabled. */
	public static final double NO_THRESHOLD = -808080.0;
	
	static public final int RED_LUT=0, BLACK_AND_WHITE_LUT=1, NO_LUT_UPDATE=2;
	static final int INVERT = 0, FILL = 1, ADD = 2, MULT = 3, AND = 4, OR = 5, XOR = 6, GAMMA = 7, LOG = 8;
	static final int BLUR_MORE=0, FIND_EDGES=1, MEDIAN_FILTER=2, MIN=3, MAX=4;
	static final double rWeight = 0.30, gWeight = 0.59, bWeight = 0.11;
	
	int fgColor = 0;
	protected int lineWidth = 1;
	protected int cx, cy; //current drawing coordinates
	protected Font font;
	protected FontMetrics fontMetrics;
	static Frame frame;
		
    ProgressBar progressBar;
    boolean pixelsModified;
	protected int width, snapshotWidth;
	protected int height, snapshotHeight;
	protected int roiX, roiY, roiWidth, roiHeight;
	protected int xMin, xMax, yMin, yMax;
	boolean newSnapshot = false; // true if pixels = snapshotPixels
	boolean killable = true; // can this processor be recreated from the AWT Image?
	int[] mask = null;
	protected ColorModel cm;
	protected ImageProducer imageSource;
	protected boolean interpolate = false;
	protected double minThreshold=NO_THRESHOLD, maxThreshold=NO_THRESHOLD;
	protected int histogramSize = 256;
	protected float[] cTable;
	
	protected void showProgress(double percentDone) {
		if (progressBar!=null)
        	progressBar.show(percentDone);
	}

	protected void hideProgress() {
		showProgress(1.0);
		newSnapshot = false;
	}
		
	public int getWidth() {
		return width;
	}
	
	public int getHeight() {
		return height;
	}
	
	/** Sets the color model. Must be an IndexColorModel for
		all processors except the ColorProcessor. */
	public void setColorModel(ColorModel cm) {
		if (!(this instanceof ColorProcessor) && !(cm instanceof IndexColorModel))
			throw new IllegalArgumentException("Must be IndexColorModel");
		this.cm = cm;
		imageSource = null;
	}

	/** Returns this processors color model. For ByteProcessors,
 		this is the base color model, not the one that may
		have been modified by setMinAndMax(). */
	public ColorModel getColorModel() {
		return cm;
	}

	/** Sets the foreground color. */
	public abstract void setColor(Color color);

	/** Obsolete (use setValue) */
	public void setColor(int value) {
		fgColor = value;
	}

	/** Sets the default fill/draw value. */
	public abstract void setValue(double value);

	/** Returns the smallest displayed pixel value. */
	public abstract double getMin();

	/** Returns the largest displayed pixel value. */
	public abstract double getMax();

	/** Map the pixels in this image from min-max to 0-255. */
	public abstract void setMinAndMax(double min, double max);

	/** For short and float images, recalculates the min and max
		image values needed to correctly display the image. For
		ByteProcessors, resets the LUT.*/
	public void resetMinAndMax() {}

	/** Sets the min and max threshold levels. */
	public void setThreshold(double minThreshold, double maxThreshold, int lutUpdate) {}

	/** Returns the minimum threshold. */
	public double getMinThreshold() {
		return minThreshold;
	}

	/** Returns the threshold maximum. */
	public double getMaxThreshold() {
		return maxThreshold;
	}

	public void setRoi(Rectangle roi) {
		if (roi==null)
			setRoi(0, 0, width, height);
		else
			setRoi(roi.x, roi.y, roi.width, roi.height);
	}

	public void setRoi(int x, int y, int rwidth, int rheight) {
		//find intersection of roi and this image
		roiX = Math.max(0, x);
		roiWidth = Math.min(width, x+rwidth)-roiX;
		roiY = Math.max(0, y);
		roiHeight = Math.min(height, y+rheight)-roiY;
		//setup limits for 3x3 filters
		xMin = Math.max(roiX, 1);
		xMax = Math.min(roiX + roiWidth - 1, width - 2);
		yMin = Math.max(roiY, 1);
		yMax = Math.min(roiY + roiHeight - 1, height - 2);
		mask = null;
	}

	public Rectangle getRoi() {
		return new Rectangle(roiX, roiY, roiWidth, roiHeight);
	}

	public void setMask(int[] mask) {
		this.mask = mask;
		if (this.mask==null)
			this.mask = null;
	}

	/** For images with irregular ROIs, returns a mask, otherwise, 
		returns null. Mask pixels have a value of BLACK. */
	public int[] getMask() {
		return mask;
	}

	/** Assigns a progress bar to this processor. Set 'pb' to
		null to disable the progress bar. */
	public void setProgressBar(ProgressBar pb) {
		progressBar = pb;
	}

	/** Setting 'interpolate' true causes scale(), resize()
		and rotate() to do bilinear interpolation. */
	public void setInterpolate(boolean interpolate) {
		this.interpolate = interpolate;
	}

	public boolean isKillable() {
		return killable;
	}

	private void process(int op, double value) {
		double SCALE = 255.0/Math.log(255.0);
		int v;
		
		int[] lut = new int[256];
		for (int i=0; i<256; i++) {
			switch(op) {
				case INVERT:
					v = 255 - i;
					break;
				case FILL:
					v = fgColor;
					break;
				case ADD:
					v = i + (int)value;
					break;
				case MULT:
					v = (int)Math.round(i * value);
					break;
				case AND:
					v = i & (int)value;
					break;
				case OR:
					v = i | (int)value;
					break;
				case XOR:
					v = i ^ (int)value;
					break;
				case GAMMA:
					v = (int)(Math.exp(Math.log(i/255.0)*value)*255.0);
					break;
				case LOG:
					if (i==0)
						v = 0;
					else
						v = (int)(Math.log(i) * SCALE);
					break;
				 default:
				 	v = i;
			}
			if (v < 0)
				v = 0;
			if (v > 255)
				v = 255;
			lut[i] = v;
		}
		applyTable(lut);
    }

	public double[] getLine(int x1, int y1, int x2, int y2) {
		int dx = x2-x1+1;
		int dy = y2-y1+1;
		int n = (int)Math.round(Math.sqrt(dx*dx + dy*dy));
		double[] data = new double[n];
		double xinc = (double)dx/n;
		double yinc = (double)dy/n;
		double rx = x1;
		double ry = y1;
		int xx, yy;
		for (int i=0; i<n; i++) {
			data[i] = getPixelValue((int)(rx+0.5), (int)(ry+0.5));
			rx += xinc;
			ry += yinc;
		}
		return data;
	}
	
	/** Returns the pixels along the horizontal line starting at (x,y). */
	public void getRow(int x, int y, int[] data, int length) {
		//if (x>=0 && (x+length)<=width && y>=0 && y<height)
		//	((ShortProcessor)this).getRow2(x, y, data, length);
		//else
			for (int i=0; i<length; i++)
				data[i] = getPixel(x++, y);
	}

	/** Returns the pixels down the column starting at (x,y). */
	public void getColumn(int x, int y, int[] data, int length) {
		for (int i=0; i<length; i++)
			data[i] = getPixel(x, y++);
	}

	/** Inserts the pixels contained in 'data' into a 
		horizontal line starting at (x,y). */
	public void putRow(int x, int y, int[] data, int length) {
		for (int i=0; i<length; i++)
			putPixel(x++, y, data[i]);
	}
	
	/** Inserts the pixels contained in 'data' into a 
		column starting at (x,y). */
	public void putColumn(int x, int y, int[] data, int length) {
		//if (x>=0 && x<width && y>=0 && (y+length)<=height)
		//	((ShortProcessor)this).putColumn2(x, y, data, length);
		//else 
			for (int i=0; i<length; i++)
				putPixel(x, y++, data[i]);
	}

	/** Sets the current drawing location. */
	public void moveTo(int x, int y) {
		cx = x;
		cy = y;
	}
	
	/** Sets the line width. */
	public void setLineWidth(int width) {
		lineWidth = width;
		if (lineWidth<1) lineWidth = 1;
	}
		
	/** Draws a line from the current drawing location to (x,y). */
	public void lineTo(int x2, int y2) {
		int dx = x2-cx;
		int dy = y2-cy;
		int absdx = dx>=0?dx:-dx;
		int absdy = dy>=0?dy:-dy;
		int n = absdy>absdx?absdy:absdx;
		double xinc = (double)dx/n;
		double yinc = (double)dy/n;
		double x = cx+0.5;
		double y = cy+0.5;
		n++;
		do {
			if (lineWidth==1)
				drawPixel((int)x, (int)y);
			else if (lineWidth==2)
				drawDot2((int)x, (int)y);
			else
				drawDot((int)x, (int)y);
			x += xinc;
			y += yinc;
		} while (--n>0);
		cx = x2; cy = y2;
	}
		
	/** Draws 2x2 dot in the current color. */
	public void drawDot2(int x, int y) {
		drawPixel(x, y);
		drawPixel(x-1, y);
		drawPixel(x, y-1);
		drawPixel(x-1, y-1);
	}
		
	/** Draws a dot using the current line width and color. */
	public void drawDot(int xcenter, int ycenter) {
		int r = lineWidth/2;
		int r2 = r*r+1;
		for (int x=-r; x<=r; x++)
			for (int y=-r; y<=r; y++)
				if ((x*x+y*y)<=r2)
					drawPixel(xcenter+x, ycenter+y);
	}

	/** Draws a string at the current location. */
	public void drawString(String s) {
		Color c;
		if (this instanceof ByteProcessor) {
			if (cm==null) makeDefaultColorModel();
			c = new Color(cm.getRGB(fgColor));
		} else if (this instanceof ColorProcessor)
			c = new Color(fgColor);
		else
			c = Color.black;
		Color background = Color.white;
		if ((c.getRGB()&0xffffff)==0xffffff)
			background = Color.black;
		if (frame==null)
			{frame = new Frame(); frame.pack();}
		frame.setBackground(background);
		if (font==null)
			font = new Font("SansSerif", Font.PLAIN, 9);
		if (fontMetrics==null)
			fontMetrics = frame.getFontMetrics(font);
		int w =  fontMetrics.stringWidth(s);
		int h =  fontMetrics.getHeight();
		Image img = frame.createImage(w, h);
		Graphics g = img.getGraphics();
		g.setColor(c);
		FontMetrics metrics = g.getFontMetrics(font);
		int fontHeight = metrics.getHeight();
		int descent = metrics.getDescent();
		g.setFont(font);
		g.drawString(s, 0, h-descent);
		g.dispose();
		ImageProcessor ip = new ColorProcessor(img);
		Blitter blitter;
		if (this instanceof ByteProcessor)
			blitter = new ByteBlitter((ByteProcessor)this);
		else if (this instanceof ColorProcessor)
			blitter = new ColorBlitter((ColorProcessor)this);
		else
			return;
		blitter.setTransparentColor(background);
		blitter.copyBits(ip, cx, cy-h, Blitter.COPY_TRANSPARENT);
		cy += h;
	}
	/** Sets the font used by drawString(). */
	public void setFont(Font font) {
		this.font = font;
		fontMetrics	= null;
	}
	
	/** Returns the width in pixels of the specified string. */
	public int getStringWidth(String s) {
		if (font==null)
			font = new Font("SansSerif", Font.PLAIN, 9);
		if (fontMetrics==null) {
			if (frame==null)
				{frame = new Frame(); frame.pack();}
			fontMetrics = frame.getFontMetrics(font);
		}
		return fontMetrics.stringWidth(s);
	}

	/** Returns the ImageProducer used by the image returned by createImage(). */
	public ImageProducer getImageSource() {
		return imageSource;
	}
		
	/** Replaces each pixel with the 3x3 neighborhood mean. */
	public void smooth() {
		filter(BLUR_MORE);
	}

	/** Sharpens the image using a 3x3 convolution kernel. */
	public void sharpen() {
		int[] kernel = {-1, -1, -1,
		                -1, 12, -1,
		                -1, -1, -1};
		convolve3x3(kernel);
	}
	
	/** This is a Sobel edge detector. */
	public void findEdges() {
		filter(FIND_EDGES);
	}

	public void flipVertical() {
		int[] row1 = new int[roiWidth];
		int[] row2 = new int[roiWidth];
		for (int y=0; y<roiHeight/2; y++) {
			getRow(roiX, roiY+y, row1, roiWidth);
			getRow(roiX, roiY+roiHeight-y-1, row2, roiWidth);
			putRow(roiX, roiY+y, row2, roiWidth);
			putRow(roiX, roiY+roiHeight-y-1, row1, roiWidth);
		}
		newSnapshot = false;
	}

	public void flipHorizontal() {
		int[] col1 = new int[roiHeight];
		int[] col2 = new int[roiHeight];
		for (int x=0; x<roiWidth/2; x++) {
			getColumn(roiX+x, roiY, col1, roiHeight);
			getColumn(roiX+roiWidth-x-1, roiY, col2, roiHeight);
			putColumn(roiX+x, roiY, col2, roiHeight);
			putColumn(roiX+roiWidth-x-1, roiY, col1, roiHeight);
		}
		newSnapshot = false;
	}

	/** Rotates the image 90 degrees clockwise. */
	public ImageProcessor rotateRight() {
		int width2 = height;
		int height2 = width;
        ImageProcessor ip2 = createProcessor(width2, height2);
		int[] arow = new int[width];
		for (int row=0; row<height; row++) {
			getRow(0, row, arow, width);
			ip2.putColumn(width2-row-1, 0, arow, height2);
		}
        return ip2;
	}
	
	/** Rotates the image 90 degrees counter-clockwise. */
	public ImageProcessor rotateLeft() {
		int width2 = height;
		int height2 = width;
        ImageProcessor ip2 = createProcessor(width2, height2);
		int[] arow = new int[width];
		int[] arow2 = new int[width];
		for (int row=0; row<height; row++) {
			getRow(0, row, arow, width);
			for (int i=0; i<width; i++) {
				arow2[i] = arow[width-i-1];
			}
			ip2.putColumn(row, 0, arow2, height2);
		}
        return ip2;
	}

	/** Inserts the image contained in 'ip' at (xloc, yloc). */
	public void insert(ImageProcessor ip, int xloc, int yloc) {
		copyBits(ip, xloc, yloc, Blitter.COPY);
	}
		
	protected void makeDefaultColorModel() {
		byte[] rLUT = new byte[256];
		byte[] gLUT = new byte[256];
		byte[] bLUT = new byte[256];
		for(int i=0; i<256; i++) {
			rLUT[i]=(byte)i;
			gLUT[i]=(byte)i;
			bLUT[i]=(byte)i;
		}
		cm = new IndexColorModel(8, 256, rLUT, gLUT, bLUT);
	}

	public void invertLut() {
		if (this instanceof ColorProcessor)
			throw new IllegalArgumentException("RGB images do not use LUTs");
		if (cm==null)
			makeDefaultColorModel();
    	IndexColorModel icm = (IndexColorModel)cm;
		int mapSize = icm.getMapSize();
		byte[] reds = new byte[mapSize];
		byte[] greens = new byte[mapSize];
		byte[] blues = new byte[mapSize];	
		byte[] reds2 = new byte[mapSize];
		byte[] greens2 = new byte[mapSize];
		byte[] blues2 = new byte[mapSize];	
		icm.getReds(reds); 
		icm.getGreens(greens); 
		icm.getBlues(blues);
		for (int i=0; i<mapSize; i++) {
			reds2[i] = (byte)(reds[mapSize-i-1]&255);
			greens2[i] = (byte)(greens[mapSize-i-1]&255);
			blues2[i] = (byte)(blues[mapSize-i-1]&255);
		}
		cm = new IndexColorModel(8, mapSize, reds2, greens2, blues2); 
		imageSource = null;
	}

	/** Returns the LUT index that's the best match for this color. */
	public int getBestIndex(Color c) {
		if (this instanceof ColorProcessor)
			throw new IllegalArgumentException("IndexColorModel required");
		if (cm==null)
			makeDefaultColorModel();
    	IndexColorModel icm = (IndexColorModel)cm;
		int mapSize = icm.getMapSize();
		byte[] rLUT = new byte[mapSize];
    	byte[] gLUT = new byte[mapSize];
		byte[] bLUT = new byte[mapSize];
    	icm.getReds(rLUT); 
    	icm.getGreens(gLUT); 
    	icm.getBlues(bLUT); 
		int minDistance = Integer.MAX_VALUE;
		int distance;
		int minIndex = 0;
		int r1=c.getRed();
		int g1=c.getGreen();
		int b1=c.getBlue();
		int r2,b2,g2;
    	for (int i=0; i<mapSize; i++) {
			r2 = rLUT[i]&0xff; g2 = gLUT[i]&0xff; b2 = bLUT[i]&0xff;
    		distance = (r2-r1)*(r2-r1)+(g2-g1)*(g2-g1)+(b2-b1)*(b2-b1);
			//ij.IJ.write(i+" "+minIndex+" "+distance+" "+(rLUT[i]&255));
    		if (distance<minDistance) {
    			minDistance = distance;
    			minIndex = i;
    		}
    		if (minDistance==0.0)
    			break;
    	}
    	return minIndex;
	}

	public String toString() {
		return ("width="+width+", height="+height+", min="+getMin()+", max="+getMax()+", v="+getPixel(0,0));
	}

	public void fill() {
		if (mask!=null) {
			fill(mask);
			mask = null;
		} else
			process(FILL, 0.0);
	}

	/** Only implemented by ShortProcessor. */
	public boolean isUnsigned() {
		return false;
	}
	
	public void setCalibrationTable(float[] cTable) {
		this.cTable = cTable;
	}

	public void setHistogramSize(int size) {
		histogramSize = size;
	}

	public int getHistogramSize() {
		return histogramSize;
	}

	public abstract int getPixel(int x, int y);
	public abstract void putPixel(int x, int y, int value);
	public abstract float getPixelValue(int x, int y);
	public abstract void drawPixel(int x, int y);
	public abstract Object getPixels();
	public abstract Object getPixelsCopy();
	public abstract void setPixels(Object pixels);
	public abstract void copyBits(ImageProcessor ip, int xloc, int yloc, int mode);

	public abstract void applyTable(int[] lut);

	public void invert() {process(INVERT, 0.0);}
	public void add(int value) {process(ADD, value);}
	public void multiply(double value) {process(MULT, value);}
	public void and(int value) {process(AND, value);}
	public void or(int value) {process(OR, value);}
	public void xor(int value) {process(XOR, value);}
	public void gamma(double value) {process(GAMMA, value);}
	public void log() {process(LOG, 0.0);}

	public abstract Image createImage();
	public abstract ImageProcessor createProcessor(int width, int height);
	public abstract void snapshot();
	public abstract void reset();
	public abstract void reset(int[] mask);
	abstract void fill(int[] mask);
	public abstract void convolve3x3(int[] kernel);
	public abstract void filter(int type);
	public abstract void medianFilter();
    public abstract void noise(double range);
	public abstract ImageProcessor crop();
	public abstract void scale(double xScale, double yScale);
	public abstract ImageProcessor resize(int dstWidth, int dstHeight);
  	public abstract void rotate(double angle);
	public abstract void autoThreshold();
	public abstract int[] getHistogram();
	public abstract void erode();
	public abstract void dilate();

}

